/* $Id: Literal.java,v 1.6 2006/03/24 18:20:23 gsson Exp $ */

/*
 * Copyright (c) 2006 Henrik Gustafsson <henrik.gustafsson@fnord.se>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package se.fnord.ecalc.parser;

public class Literal implements IValue, IOperators {
	protected int preferredType = IValue.TYPE_UNDEFINED;
	protected double dvalue = 0.0;
	protected long ivalue = 0;
	
	
	public Literal() {
	
	}
	public Literal(double d) {
		preferredType = IValue.TYPE_DOUBLE;
		dvalue = d;
	}
	public Literal(int i) {
		preferredType = IValue.TYPE_LONG;
		ivalue = i;
	}
	
	public Literal(IValue init) throws ExpressionException {
		fromIValue(init);
	}
	
	public double toDouble() throws ExpressionException {
		if (preferredType == IValue.TYPE_UNDEFINED)
			throw new ExpressionException("%se.fnord.ecalc.exceptions.expressionexception.undefined_variable");
		if (preferredType == IValue.TYPE_LONG) {
			return ivalue;
		}
		else {
			return dvalue;
		}
	}

	public long toLong() throws ExpressionException {
		if (preferredType == IValue.TYPE_UNDEFINED)
			throw new ExpressionException("%se.fnord.ecalc.exceptions.expressionexception.undefined_variable");		
		if (preferredType == IValue.TYPE_LONG) {
			return ivalue;
		}
		else {
			return (long)dvalue;
		}
	}

	public void fromDouble(double value) {
		preferredType = IValue.TYPE_DOUBLE;
		dvalue = value;
	}

	public void fromLong(long value) {
		preferredType = IValue.TYPE_LONG;
		ivalue = value;
	}

	public int getPreferredType() throws ExpressionException {
		return preferredType;
	}

	public void fromIValue(IValue value) throws ExpressionException {
			
		if (value.getPreferredType() == IValue.TYPE_LONG) {
			this.fromLong(value.toLong());
		}
		else {
			this.fromDouble(value.toDouble());			
		}
	}
	
	

	public IValue add(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			this.fromDouble(this.toDouble() + v.toDouble());
		}
		else {
			this.fromLong(this.toLong() + v.toLong());
		}
		return this;
	}

	public IValue subtract(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			this.fromDouble(this.toDouble() - v.toDouble());
		}
		else {
			this.fromLong(this.toLong() - v.toLong());
		}
		return this;
	}

	public IValue multiply(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			this.fromDouble(this.toDouble() * v.toDouble());
		}
		else {
			this.fromLong(this.toLong() * v.toLong());
		}
		return this;
	}

	public IValue divide(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			if (v.toDouble() == 0.0)
				 throw new ExpressionException("%se.fnord.ecalc.exceptions.expressionexception.divide_by_zero");
			this.fromDouble(this.toDouble() / v.toDouble());
		}
		else {
			if (v.toLong() == 0)
				 throw new ExpressionException("%se.fnord.ecalc.exceptions.expressionexception.divide_by_zero");
			this.fromLong(this.toLong() / v.toLong());
		}
		return this;
	}

	public IValue modulus(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			this.fromDouble(Math.IEEEremainder(this.toDouble(), v.toDouble()));
		}
		else {
			this.fromLong(this.toLong() % v.toLong());
		}
		return this;
	}

	public IValue power(IValue v) throws ExpressionException {
		if (preferredType == IValue.TYPE_DOUBLE || v.getPreferredType() == IValue.TYPE_DOUBLE) {
			this.fromDouble(Math.pow(this.toDouble(), v.toDouble()));
		}
		else {
			this.fromLong(Math.round(Math.pow(this.toLong(), v.toLong())));
		}
		return this;
	}

	public IValue negate() {
		if (preferredType == IValue.TYPE_DOUBLE)
			dvalue = - dvalue;
		else
			ivalue = - ivalue;
		return this;
	}
	
	public String toString() {
		if (preferredType == IValue.TYPE_DOUBLE)
			return Double.toString(dvalue);
		else
			return Long.toString(ivalue);
		
	}

	public Literal ensureType(int type) throws ExpressionException {
		if (this.preferredType == type || type == IValue.TYPE_AUTO)
			return new Literal(this);
		else {
			Literal v  = new Literal();
			if (type == IValue.TYPE_DOUBLE) {
				v.fromDouble(toDouble());
			}
			else {
				v.fromLong(toLong());
			}
			return v;
		}
	}

}
